#include "stdafx.h"
#include "util.h"
#include "cwaitcur.h"
#include "resource.h"
#include "const.hxx"
#include "dbsync.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//		class CDBSyncException() implementation
//
//      can be thrown from the DoSync() member of class CDBSync
//
//      the most likely error is a failure to connect to the CDBSync::m_strDSN datasource (it is wrong in the DBSYNC.INI file, probably)
//
IMPLEMENT_DYNAMIC(CDBSyncException, CException)

CDBSyncException::CDBSyncException(int nErrCode)
{
	m_nErrCode = nErrCode;
	if(m_nErrCode >= ERR_DBSYNC_BADPARAMS && m_nErrCode <= ERR_DBSYNC_OPENINGUSERTRANSTBL)
	{
		VERIFY(m_strError.LoadString(m_nErrCode));
	}
	else
		m_strError = "CDBSync Undefined Exception";
}
//exception cleanup
CDBSyncException::~CDBSyncException()
{
	m_strError.Empty();
}

void AFXAPI ThrowDBSyncException(int nErrCode)
{
	CDBSyncException* e = new CDBSyncException(nErrCode);
	THROW(e);
}

//local only
BOOL match_quotes(LPSTR lpsz);

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//		class CDBSync() implementation
//
//
//
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
IMPLEMENT_DYNAMIC(CDBSync, CObject)

//constructor
CDBSync::CDBSync(LPCSTR lpcstrFileName /*= NULL*/, LPCSTR lpcstrOrigFileTitle /*= NULL*/, LPCSTR lpcstrDSN /*= NULL*/)
{
	m_strFileName = "";
	m_strOrigFileTitle = "";
	m_strDSN = "";
	m_pDb = NULL;
	m_pAddTransSet = NULL;
	m_pFile = NULL;
	m_szLineBuf = NULL;
	m_szCopyBuf = NULL;
	m_bSaveLocalCopy = FALSE;
	m_pStatusDlg = NULL;
	m_bUserAbort = FALSE;	
	m_strSenderMailName = "";
	
	//always show messages when we are debugging
	#ifdef _DEBUG
		m_bShowErrorMessages = TRUE;
	#else
		m_bShowErrorMessages = ::GetPrivateProfileInt(SZ_INISEC_DBSYNC, SZ_INIKEY_SHOWERRORMESSAGES, 0, SZ_INIFILE);
	#endif //_DEBUG
	
	Init(lpcstrFileName, lpcstrOrigFileTitle, lpcstrDSN);
	
}

//destructor
CDBSync::~CDBSync()
{
	
	//destroy the status dialog, 
	
	if(m_pStatusDlg && ::IsWindow(m_pStatusDlg->m_hWnd))
	{
		m_pStatusDlg->DestroyWindow();
		delete m_pStatusDlg;
	}	
	m_pStatusDlg = NULL; //the object is deleted
	
	if(m_pDb){
	
		if(m_pAddTransSet){
			if(m_pAddTransSet->IsOpen())
				m_pAddTransSet->Close();
				
			delete m_pAddTransSet;
			m_pAddTransSet = NULL;
					
		}
	
		if(m_pDb->IsOpen())
			m_pDb->Close();
	
		delete m_pDb;		
		m_pDb = NULL;
	
	}	
	
		
	if(m_pFile){
		
		if(m_pFile->IsOpen())
			m_pFile->Close();
			
		delete m_pFile;	
	}
	
	if(m_szLineBuf)
		delete m_szLineBuf;
		
	if(m_szCopyBuf)
		delete m_szCopyBuf;
	
	
	//clear everything
	m_pFile = NULL;
	m_pDb = NULL;
	m_strFileName.Empty();
	m_strOrigFileTitle.Empty();
	m_strDSN.Empty();
	m_strSenderMailName.Empty();
	m_szLineBuf = NULL;
	m_szCopyBuf = NULL;
	m_pAddTransSet = NULL;
	m_pStatusDlg = NULL;	
	m_bUserAbort = FALSE;
	
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//			CDBSync::Init
//
//	        LPCSTR lpcstrFileName			name of the .DBX file
//			LPCSTR lpcstrDSN				  name of the datasource (pointing to SQL server or DBXDATA.MDB)
//
//			Called from the constructor
//
//			Creates the text file object (CWinFile), database object (CDatabase) and
//			the string buffers for reading the SQL statements out of the file.
//
//
//          no return value, may throw an exception, (memory, ????)
//			if an exception is thrown, then the caller has to catch it
//
BOOL CDBSync::Init(LPCSTR lpcstrFileName, LPCSTR lpcstrOrigFileTitle, LPCSTR lpcstrDSN)
{
 	if(lpcstrFileName)
 		m_strFileName = lpcstrFileName;
 	if(lpcstrDSN)
 		m_strDSN = lpcstrDSN;
 	if(lpcstrOrigFileTitle)
 		m_strOrigFileTitle = lpcstrOrigFileTitle;
 		
 	if(lpcstrFileName && m_strOrigFileTitle.IsEmpty())
 	{
		::GetFileTitle(m_strFileName, m_strOrigFileTitle.GetBuffer(256), 256);
		m_strOrigFileTitle.ReleaseBuffer();
		TRACE1("CDBSync::m_strOrigFileTitle == %s\n", (LPCSTR)m_strOrigFileTitle);
	}		
	 		
 	TRY
 	{
 		m_pDb = new CDBSyncDatabase; 	 	
 		m_pFile = new CWinFile;
 		
 		m_szLineBuf = new char[MAX_SQL_STMT_SIZE];
 		memset(m_szLineBuf, 0, MAX_SQL_STMT_SIZE);
 		
 		m_szCopyBuf = new char[MAX_SQL_STMT_SIZE];
 		memset(m_szCopyBuf, 0, MAX_SQL_STMT_SIZE);
 		
 	}
 	CATCH_ALL(e)
 	{
 		//nothing to do but run 		
 		THROW_LAST(); //out to the caller
 	}
 	END_CATCH_ALL
 	
 	return TRUE;
 	
}




////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//			CDBSync::OpenDatabase()
//			
//
//			Connects to a datasource
//
//
//			returns TRUE (opened OK) or FALSE (bomb)
//			ALSO, the bomb (error) will throw and catch its own exception (CDBException)
//			and possibly display an error message
//
BOOL CDBSync::OpenDatabase()
{
	if(NULL == m_pDb || NULL == m_pFile)
		return FALSE;
		
	CString strConnect, strDB;
	
	TRY
	{
	
		TRACE1("m_strDSN: '%s'\n", (LPCSTR)m_strDSN);		
				
		if(m_strDSN.IsEmpty())
		{
			strConnect = "ODBC;";
			return m_pDb->Open(NULL, FALSE, FALSE, strConnect);
		}
		else{
			
			if(m_bSaveLocalCopy){	//Should be SQL server
				
				strConnect = "ODBC;DSN=";
				strConnect += m_strDSN;
				strConnect += ";DATABASE=";
				
				//	get the name of the database (to corp)
				//	this is just in case someone uses the ODBC name in another app and changes the database that it points to
				//	most likely this would have been some developer				
				::GetPrivateProfileString(SZ_INISEC_DBSYNC, SZ_INIKEY_DBSYNCDATABASENAME, "", strDB.GetBuffer(80), 80, SZ_INIFILE);
				strDB.ReleaseBuffer();
				
				strConnect += strDB;
				strConnect += ";";
				
				//add the user id/ password	(SQL server requires this)
				strConnect += SZ_DBSYNC_UID;
				strConnect += ";";
				strConnect += SZ_DBSYNC_PWD;
				
				if(m_pDb->Open(NULL, FALSE, FALSE, strConnect))
					return TRUE;
				else
				{
					strConnect = "ODBC;";
					return m_pDb->Open(NULL, FALSE, FALSE, strConnect);
				}
			}
			else			
			{
				//THIS is probably the Jet database, so SECURITY is ignored (shame),
				//however, this probably should be changed in a "real" implementation
				
				//should be applying to the remote db, don't use security
				return m_pDb->Open(m_strDSN);		
			}
		}
	}
	CATCH(CDBException, e)
	{
		strConnect.Empty();
		strDB.Empty();
		
		DoErrorMsg(e);	
	}
	AND_CATCH(CException, e)
	{
		//something exceptionally dreadful happened, run away
		TRACE("Unhandled excption in DBSYNC.CPP\n");
	}
	END_CATCH
	
	return FALSE;
		
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//			CDBSync::LogAsApplied()
//
//			int iRowsRead							number of rows read
//			int iRowsApplied						number of rows applied (without error)
//
//
//
//		Logs a row to the "APPLIED" table in the recieving databse 
//		This is the Table that stores information about the applied files (SQL statements) for the application to read 
//		in order to send confirmation messages back to the sender
//
//		YOU MUST REBUILD THIS .DLL AND CHANGE THE FOLLOWING CODE (particularly the szInsertSQL)
//		IF YOU CHANGE THE NAME OF THIS TABLE (the dbx_applied table)
//
//
static char BASED_CODE szInsertSQL[] = "INSERT INTO dbx_applied (LogDate, RowsRead, RowsApplied, UserAbort, DBXFileName, SentBy, Reply) ";
static char BASED_CODE szLogAsAppliedMsg[] = "Logging succuss...";
BOOL CDBSync::LogAsApplied(int iRowsRead, int iRowsApplied)
{
	
	m_pStatusDlg->SetStatusMessage(szLogAsAppliedMsg, "");
	
	//builds a SQL string (the Insert part) 
	//and adds it to the end of the above szInsertSQL string 
	char sz[18];
	CTime t = CTime::GetCurrentTime();
	
	static char BASED_CODE szQ[] = "'";
	static char BASED_CODE szComma[] = ", ";
	
	CString strValues = szInsertSQL;
	strValues += " VALUES('";
	strValues += FormatCTime(&t);
	strValues += szQ;
	strValues += szComma;
	
	strValues += itoa(iRowsRead, sz, 10);
	strValues += szComma;
	
	strValues += itoa(iRowsApplied, sz, 10);
	strValues += szComma;
	
	if(m_bUserAbort)
		strValues += "-1";
	else
		strValues += "0";
	strValues += szComma;
	
	strValues += szQ;
	
	char szFileTitle[16];
	::GetFileTitle(m_strFileName, szFileTitle, sizeof(szFileTitle));
	strValues += szFileTitle;
	strValues += szQ;
	strValues += szComma;
	
	strValues += szQ;
	strValues += m_strSenderMailName;
	strValues += "', 0)";	
	
	TRY
	{
		return ExecuteSQL(strValues);	
	}
	CATCH(CDBException, e)
	{
		TRACE("CDBSync::LogAsApplied()->Error logging success\n");
		DoErrorMsg(e);
	}
	AND_CATCH(CException, e)
	{
	    TRACE("Unhandled excption in DBSYNC.CPP - LogAsApplied()\n");
	}
	END_CATCH
	
		
	return FALSE;
		
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//			CDBSync::ExecuteSQL()
//
//			LPCSTR lpszSQL					the SQL statement (INSERT, UPDATE, DELETE)
//
//
//
//			Executes a SQL statement
//			
//			returns TRUE (successful SQL) or FALSE (SQL error)
//
//			May throw a CDBException
//
//
//			NOTICE that this is actually an overridden form (copy) of the MFC CDatabase::ExecuteSQL()
//			ALSO, the CDBException thrown is constructed IN THIS FUNCTION, in order to drop the 
//			active statement before we THROW out
//
BOOL CDBSyncDatabase::ExecuteSQL(LPCSTR lpszSQL)
{
	BOOL bSuccess = FALSE;
	RETCODE nRetCode;

	ASSERT_VALID(this);
	ASSERT(AfxIsValidString(lpszSQL));
	// Can't close till all pending Async operations have completed
	ASSERT(!InWaitForDataSource());

	ASSERT(m_hstmt == SQL_NULL_HSTMT);
	AFX_SQL_SYNC(::SQLAllocStmt(m_hdbc, &m_hstmt));
	if (!Check(nRetCode))
		ThrowDBException(nRetCode);

	TRY
	{
		OnSetOptions(m_hstmt);

		AFX_SQL_ASYNC(this, ::SQLExecDirect(m_hstmt, (UCHAR FAR*)lpszSQL, SQL_NTS));
		if (Check(nRetCode))	//everything worked
		{
			do
			{
				UINT nResultColumns;

				AFX_SQL_ASYNC(this, ::SQLNumResultCols(m_hstmt, (short FAR*)&nResultColumns));
				if (nResultColumns != 0)
				{
					do
					{
						AFX_SQL_ASYNC(this, ::SQLFetch(m_hstmt));
					} while (Check(nRetCode) && nRetCode != SQL_NO_DATA_FOUND);
				}
				AFX_SQL_ASYNC(this, ::SQLMoreResults(m_hstmt));
			} while (Check(nRetCode) && nRetCode != SQL_NO_DATA_FOUND);
		}
		else
		{
			//create and throw the exception HERE, because it would be best if we dropped the 
			//HSTMT before moving on			
			//this code was taken (and modified) from DBCORE.CPP (AfxThrowDBException)
			CDBException* pException = new CDBException(nRetCode);
			if (nRetCode == SQL_ERROR)
			{
				pException->BuildErrorString(this, m_hstmt);
			}
			else if (nRetCode > AFX_SQL_ERROR && nRetCode < AFX_SQL_ERROR_MAX)
			{
				VERIFY(pException->m_strError.LoadString(AFX_IDP_SQL_FIRST+(nRetCode-AFX_SQL_ERROR)));
				TRACE1("%s\n", (LPCSTR)pException->m_strError);
			}
			
			//now drop the statement and set it back to null
			AFX_SQL_SYNC(::SQLFreeStmt(m_hstmt, SQL_DROP));
			m_hstmt = SQL_NULL_HSTMT;
		
			THROW(pException);
			
			return FALSE;	//we will never get here
		}

		bSuccess = (nRetCode == SQL_NO_DATA_FOUND);
	}
	CATCH_ALL(e)
	{
		::SQLCancel(m_hstmt);
		AFX_SQL_SYNC(::SQLFreeStmt(m_hstmt, SQL_DROP));
		m_hstmt = SQL_NULL_HSTMT;
		THROW_LAST();
	}
	END_CATCH_ALL

	AFX_SQL_SYNC(::SQLFreeStmt(m_hstmt, SQL_DROP));
	m_hstmt = SQL_NULL_HSTMT;
	return bSuccess;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//			CDBSync::ExecuteSQL()
//			LPCSTR lpcstrSQL				the SQL statement
//
//
//			this is a wrapper for the CDBSyncDatabase::ExecuteSQL() function
//			
//			This function catches any exception that the previous Execute may throw and
//			possibly displays and error message
//
BOOL CDBSync::ExecuteSQL(LPCSTR lpcstrSQL)
{
	
	TRY
	{
		TRACE1("CDBSync::ExecuteSQL::lpcstrSQL:  %s\n", lpcstrSQL);
		return m_pDb->ExecuteSQL(lpcstrSQL);
	}
	CATCH(CDBException, e)
	{
		DoErrorMsg(e);		
	}
	AND_CATCH(CException, e)
	{
		TRACE("Unhandled excption in DBSYNC.CPP - ExecuteSQL()\n");
	}
	END_CATCH
	
	
	return FALSE;
	
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//			CDBSync::DoSync()
//
//
//
//			main entry point (called from _apply_trans)
//
//          returns TRUE if successful 
//          or FALSE  (may throw a CDBSyncException)
//			
//
BOOL CDBSync::DoSync()
{

	static char szStatus1[40], szStatus2[40];	//for the dialog messages
		
	if(NULL == m_pDb || NULL == m_pFile)
		return FALSE;
		
	if(m_strFileName.IsEmpty() || !FileExists(m_strFileName))
	{
		ThrowDBSyncException(ERR_DBSYNC_MISSINGFILE);
		return FALSE;
	}
		
	// create and show the status dialog before we try to open the database (connection times may vary)
	if(m_bSaveLocalCopy)
		m_pStatusDlg = new CStatusDlg(&m_ParentWnd, IDD_DBSYNCTOCORPSTATUS);
	else
		m_pStatusDlg = new CStatusDlg(&m_ParentWnd, IDD_DBSYNCSTATUS);
	m_pStatusDlg->SetStatusMessage("Opening...", "");
		
	//got to have a database
	if(!OpenDatabase())
	{
		m_strDSN.Empty();
		if(!OpenDatabase())
		{
	    	ThrowDBSyncException(ERR_DBSYNC_OPENDATABASE);
			return FALSE;
		}
		
	}
		
	//check the dbx_applied table to see if this file has already been successfully applied
	//(no confirmation yet)
	TRY
	{
					
		if(IsApplied(m_strOrigFileTitle, m_strSenderMailName))
			return TRUE;
	}
	CATCH(CDBException, e)
	{
		DoErrorMsg(e);
	}
	END_CATCH
				
	m_pAddTransSet = NULL;
	
	//if we are saving a local copy of the SQL statements (SQL Server only)
	//then we need to open a recordset for the .Append() stuff
	if(m_bSaveLocalCopy){
	
		TRY
		{
			
			m_pAddTransSet = new CAddTrans(m_pDb);
			if(!m_pAddTransSet->Open(CRecordset::snapshot, NULL, CRecordset::appendOnly))		//APPEND ONLY!
				return FALSE;
			
		}
		CATCH(CDBException, e)
		{
			DoErrorMsg(e);
			ThrowDBSyncException(ERR_DBSYNC_OPENINGUSERTRANSTBL);				
			return FALSE;
		}
		AND_CATCH(CException, e)
		{
		    TRACE("Exception in DoSync. Ignored.\n");
			return FALSE;
		}
		END_CATCH
	
	}		
		
	CStringArray strValueList;
	int nCols = 0;
	BOOL bSuccess = FALSE;
	BOOL bBeginTrans = FALSE;
		
	TRY
	{
	    		
		if(!m_pFile->Open(m_strFileName))
		{
			ThrowDBSyncException(ERR_DBSYNC_OPENINGFILE); //couldn't open the .DBX file.
			return FALSE;
		}
				
		//see the constructor for the creation of these 
		memset(m_szLineBuf, 0, MAX_SQL_STMT_SIZE); 
 		memset(m_szCopyBuf, 0, MAX_SQL_STMT_SIZE);
 		
		//first line of the file
		m_pFile->ReadString(m_szLineBuf, MAX_SQL_STMT_SIZE);				
		
		
		int iRow = 0;
		int iAppliedRow = 0;
				
		while(m_szLineBuf[0]){			
		
			MSG msg;
			m_bUserAbort = m_pStatusDlg->UserAbort();
			
			while (!(m_bUserAbort = m_pStatusDlg->UserAbort()) &&
				::PeekMessage(&msg, NULL, NULL, NULL, PM_NOREMOVE))
			{
				if (!AfxGetApp()->PumpMessage())
					m_bUserAbort = TRUE;
			}
			
			if(TRUE == (m_bUserAbort = m_pStatusDlg->UserAbort())){
				if(IDYES == AfxMessageBox(IDS_CONFIRMABORT, MB_YESNO | MB_ICONQUESTION))
					break;
				m_pStatusDlg->SetUserAbort(FALSE);
			}
		
			//check for line breaks in the middle of the line
			while('\0' != m_szLineBuf[0] && !match_quotes(m_szLineBuf)){
				
				if(NULL == m_pFile->ReadString(m_szCopyBuf, MAX_SQL_STMT_SIZE)) //eof?
					break;					
				
				lstrcat(m_szLineBuf, "\r\n"); //add back the crlf
				
				strncat(m_szLineBuf, m_szCopyBuf, ((MAX_SQL_STMT_SIZE - lstrlen(m_szLineBuf))-1) );				
				
				memset(m_szCopyBuf, 0, MAX_SQL_STMT_SIZE);
				
				
			}
			
			if('\0' == m_szLineBuf[0])
				break;
				
			nCols = StringToStringArray(m_szLineBuf, &strValueList, '\t');
			
			// each line should be in the format StatusUserID, CustomerID, SQL
			
			if(nCols == 3){
				
				//	Wrap each execute in a single transaction.  we want as many of these to succeed as possible
				m_pDb->BeginTrans();
				bBeginTrans = TRUE;
				bSuccess = ExecuteSQL(strValueList[2]);				
				m_pDb->CommitTrans();
				bBeginTrans = FALSE;
				
				//update the status dialog
				if(bSuccess)
				{
					//	almost everything should succeed, however, some things may have to fall through the cracks
					//	for instance, if a row has been deleted already, the next delete will fail -- this is OK
					//  or a duplicate row is (attempted) inserted.  Since the row is already in the table, the failure is not
					//  really important, as future updates will remedy any other changes
					//  If the SQL statement is "BAD" then we have a far more serious bug than could be decrypted in this
					//  function, so we (the user) should probably turn on error messages to see better what is wrong
					++iAppliedRow;
					
				}
				
				wsprintf(szStatus1, "Updates read: %i...", ++iRow);
				wsprintf(szStatus2, "Updates applied: %i...", iAppliedRow);				
				m_pStatusDlg->SetStatusMessage(szStatus1, szStatus2);
				
				//if we are saving a local copy, then insert this transaction into the dbx_applied table
				if(bSuccess && m_bSaveLocalCopy)
				{
					InsertTransaction(strValueList);
				}
								
				//fresh start
				memset(m_szLineBuf, 0, MAX_SQL_STMT_SIZE);
 		        memset(m_szCopyBuf, 0, MAX_SQL_STMT_SIZE);
				
				if(NULL == m_pFile->ReadString(m_szLineBuf, MAX_SQL_STMT_SIZE)) //eof?
					break;
			
			}//while  ('\0' != m_szLineBuf[0] && !match_quotes(m_szLineBuf))
			
			strValueList.RemoveAll();
			strValueList.FreeExtra();
		
		}	
				
		LogAsApplied(iRow, iAppliedRow);
		
		return TRUE;
	
	}
	CATCH(CDBException, e)
	{
		DoErrorMsg(e);		
		
		//no hanging transactions
		if(bBeginTrans)
			m_pDb->Rollback();
		
		bBeginTrans = FALSE;
		
		ThrowDBSyncException(ERR_DBSYNC_EXCEPTION);
		return FALSE;
	
	}
	AND_CATCH(CException, e)
	{
		TRACE("Unhandled excption in DBSYNC.CPP - ExecuteSQL()\n");
		return ERR_DBSYNC_EXCEPTION;
	}
	END_CATCH
	
	//should never get here	
	return FALSE;
			
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//			CDBSync::IsApplied
//				LPCSTR lpcstrFile			.DBX file name
//				LPCSTR lpcstrSentBy			the (mail) user who sent this file
//
//
//		looks in dbx_applied for THIS file from THIS sender,
//
//		returns TRUE if its there, FALSE if its not
//		
//		this may throw a CDBException, though this is unlikely, probably means that the dbx_applied table is missing.
//
static char BASED_CODE szSQLFmt[] = "SELECT COUNT(DBXFileName) FROM dbx_applied WHERE DBXFileName = '%s' And SentBy = '%s'";
BOOL CDBSync::IsApplied(LPCSTR lpcstrFile, LPCSTR lpcstrSentBy) const
{
	long lCount = 0L;
	HSTMT hstmt = SQL_NULL_HSTMT;
	long lNumberBytes = 0L;
	RETCODE nRetCode;
		
    while(m_pDb->InWaitForDataSource())
		m_pDb->OnWaitForDataSource(FALSE);     	
    			
	AFX_SQL_SYNC(::SQLAllocStmt(m_pDb->m_hdbc, &hstmt));
	
	CString strSQL;
	char szFileTitle[16];
	::GetFileTitle(lpcstrFile, szFileTitle, sizeof(szFileTitle));
	int nLen = (lstrlen(szSQLFmt) + lstrlen(szFileTitle) + lstrlen(lpcstrSentBy) + 4);
	wsprintf(strSQL.GetBuffer(nLen), szSQLFmt, szFileTitle, lpcstrSentBy);
	strSQL.ReleaseBuffer();
	
	TRACE1("CDBSync::IsApplied() strSQL:  %s\n", (LPCSTR)strSQL);
	
	AFX_SQL_SYNC(::SQLExecDirect(hstmt, (unsigned char far *)(LPCSTR)strSQL, SQL_NTS));
    
	if (nRetCode == SQL_SUCCESS)
	{
		AFX_SQL_SYNC(::SQLBindCol(hstmt, 1, SQL_C_LONG, (void far *)&lCount, sizeof(long), &lNumberBytes));
    	AFX_SQL_SYNC(::SQLFetch(hstmt));
    	
    	TRACE1("CDBSync::IsApplied() -> lCount: %li\n", lCount);
    	
		if(nRetCode == SQL_ERROR)
		{
		    CDBException* pException = new CDBException(nRetCode);
			if (nRetCode == SQL_ERROR)
			{
				pException->BuildErrorString(m_pDb, hstmt);
			}
			else if (nRetCode > AFX_SQL_ERROR && nRetCode < AFX_SQL_ERROR_MAX)
			{
				VERIFY(pException->m_strError.LoadString(AFX_IDP_SQL_FIRST+(nRetCode-AFX_SQL_ERROR)));
				TRACE1("%s\n", (LPCSTR)pException->m_strError);
			}
			
			//now drop the statement
			AFX_SQL_SYNC(::SQLFreeStmt(hstmt, SQL_DROP));			
		
			THROW(pException);
			
			return FALSE;	//we will never get here			
		}
		
	}
	else
	{
		::SQLCancel(hstmt);		
	}
		
	BOOL bIsApplied = (lCount > 0);
	AFX_SQL_SYNC(::SQLFreeStmt(hstmt, SQL_DROP));
	
	return bIsApplied;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
//		CDBSync::InsertTransaction()
//			CStringArray& strValueList			(reference to a string array) array of values
//				index 0 == UserID, index 1 == CustomerID, index 2 == SQL statement
//
//		Appends the dbx_user_transactions table with the contents of the current (hopefully successful) transaction
// 		from the .DBX text file
//
//		No exceptions are thrown, an error return (FALSE) is handled by the caller
//
BOOL CDBSync::InsertTransaction(CStringArray& strValueList)
{
	ASSERT(strValueList.GetSize() >= 1);
	
	TRY
	{
		m_pAddTransSet->AddNew();
		
			m_pAddTransSet->m_EventDate = CTime::GetCurrentTime();
			m_pAddTransSet->m_StatusUserID = atol(strValueList[0]);
			m_pAddTransSet->m_CustomerID = atol(strValueList[1]);
			m_pAddTransSet->m_SQL = strValueList[2];
						
		return (m_pAddTransSet->Update());
		
	}
	CATCH(CDBException, e)
	{
		DoErrorMsg(e);	
	}
	END_CATCH
	
	return FALSE;
}	

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//		CDBSync::DoErrorMsg()
//			CDBException* e					pointer to a CDBException object
//
//
//		since these messages are probably shown from within a custom message in Microsoft mail, ONLY show them
//		if the user has specified in the .INI file
//		CAUTION:  a user should not use the View/New Messages command in Mail to start the processing, 
//		under certain circumstances (in WFW) this command creates a window
//		that mail does not know seem to know about which can cause some problems when responding to the message box
//
//
void CDBSync::DoErrorMsg(CDBException* e) const
{
	TRACE1("CDBException::m_strError: %s\n", e->m_strError);
	if(m_bShowErrorMessages)
	{
		CString strMsg = e->m_strStateNativeOrigin;
		strMsg += "\r\n";
		strMsg += e->m_strError;
		::MessageBox(NULL, strMsg, "Database Error", MB_ICONHAND | MB_TASKMODAL);
	}
}






////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//		match_quotes()
//
//
//			searchs a string for an even number of quotes (ascii character 34) in a string
//			this allows for newline charaters in a line of a file
//
//				"field","field
//				","field",etc...
//
//		returns TRUE (quotes match) or FALSE (they don't)			
//
//
BOOL match_quotes(LPSTR lpsz)
{
	int i = 0;
	int q = 0;
	
	for(i = 0; lpsz[i] != 0;++i){
		if(34 == lpsz[i])
			q++;
	}
	
	return ((q % 2) == 0);	

}


//////////////////////////////////////////////////////////
//
//		static BOOL CDBSync::ApplyTransactionFile()
//				HWND hWnd													Mail main window
//				LPCSTR lpcstrSenderMailName							mail name of the sender
//				LPCSTR lpcstrFileName									actual .DBX file name
//				LPCSTR lpcstrOrigFileTitle							original file title (in case mail had to create a temp file)
//				LPCSTR lpcstrDSN										the datasource name
//				BOOL bSaveLocalCopy										true specifies that a local copy of the transactions should be saved
//
//		applies the "file" of SQL statements to the appropriate database
//		creates a CDBSync() object with the datasource and filename as initial paramaters
//	
//
//	No exceptions are thrown from this function
//
//	returns TRUE (success and no abort by the user)
// or 
// FALSE (error)
//
BOOL CDBSync::ApplyTransactionFile(HWND hWnd, 
LPCSTR lpcstrSenderMailName, LPCSTR lpcstrFileName, LPCSTR lpcstrOrigFileTitle, 
LPCSTR lpcstrDSN, BOOL bSaveLocalCopy /*= FALSE*/)
{
	int bSuccess = FALSE;
		
	if(NULL == lpcstrFileName || 0 == lstrlen(lpcstrFileName) || !FileExists(lpcstrFileName) || NULL == lpcstrDSN)
		return -1;
		
	CDBSync* pDBSync = NULL;
	TRY
	{
		//create the object and fill some of the public members
		pDBSync = new CDBSync(lpcstrFileName, lpcstrOrigFileTitle, lpcstrDSN);	//may bomb, throw CMemoryException
		
		pDBSync->m_bSaveLocalCopy = bSaveLocalCopy; //copy the successful transactions to the local file? (DBSyncToCorp messages)
		
		//if the sender is empty or null, then supply a bogus name, (this should never really happen), except when
		//we are debugging
		if(lpcstrSenderMailName && lstrlen(lpcstrSenderMailName) > 0)
			pDBSync->m_strSenderMailName = lpcstrSenderMailName;
		else
			pDBSync->m_strSenderMailName = "Unknown Sender";
			
		//since this is a .DLL and there is no real window
		//create a temporary CWnd object to point to for the "main" window
		//odbc needs a parent window in some cases
		pDBSync->m_ParentWnd.Attach(hWnd);
		AfxGetApp()->m_pMainWnd = &pDBSync->m_ParentWnd;	
		
		//DoSync may throw an exception, or return some sort of error, or return that the apply was 
		//aborted (unlikely) by the user
		if(TRUE == pDBSync->DoSync())
			bSuccess = 	!(pDBSync->UserAborted());		
		
	}
	CATCH(CDBSyncException, e)
	{
		if(!e->m_strError.IsEmpty())
			AfxMessageBox(e->m_strError, MB_ICONINFORMATION | MB_TASKMODAL);
	}
	AND_CATCH(CException, e)	//catch and ignore any other exceptions
	{
		TRACE("Unhandled exception in _apply_trans()\n");
		bSuccess = FALSE;
	}
	END_CATCH
	
	
    //and clean up
	if(pDBSync)
	{
		if(pDBSync->m_ParentWnd.m_hWnd)
		{
			pDBSync->m_ParentWnd.Detach();
			pDBSync->m_ParentWnd.DeleteTempMap();
			AfxGetApp()->m_pMainWnd = NULL;
		}
		
		delete pDBSync; //closes the database, destroys the status dialog, ...
	}

	pDBSync = NULL;
	
	return bSuccess;
	
}


//////////////////////////////////////////////////////////
//
//		ApplyTransactionFile()
//			LPCSTR lpcstrFileName			.DBX File name
//			LPCSTR lpcstrDSN					datasource name
//			BOOL bSaveLocalCopy				copy SQL statements to the local dbx_user_transactions table?
//
//	External version of the function (CDBSync::ApplyTransactionFile), 
//	this version is declared extern "c" and is visible to other applications, (such as testdll.exe)
//
//	THIS FUNCTION IS NOT USED EXCEPTION FOR DEBUGGING IN THIS IMPLEMENTATION
//	
// 		
//
#ifdef _DEBUG
int WINAPI DBX_EXPORT ApplyTransactionFile(LPCSTR lpcstrFileName, LPCSTR lpcstrDSN, BOOL bSaveLocalCopy)
{
	AfxLockTempMaps();
	int rc = CDBSync::ApplyTransactionFile(::GetActiveWindow(), "", lpcstrFileName, NULL, lpcstrDSN, bSaveLocalCopy);
	AfxUnlockTempMaps();

	#ifdef _DEBUG			
		CMemoryState memState;
		memState.DumpAllObjectsSince();		
	#endif		//_DEBUG

	return rc;
}
#endif //_DEBUG
